#include <stdint.h>
#include <stdbool.h>
#include <string.h>
#include "Config.h"
#include "stm32f107.h"

static volatile uint32_t counter;
static volatile uint32_t led_counter;

static inline void delay_ms(uint32_t ms){
  counter = 0;
  while(counter < ms);
}

static void init();
void memclr(void *buffer, size_t size);

// Регистры ULCD
#define ULCD_STATUS           (0x21)
#define ULCD_SYSFONT          (0x24)
#define ULCD_WIDTH            (0x2A)
#define ULCD_HEIGHT           (0x2B)
#define ULCD_INTERRUPT        (0x30)
#define ULCD_CURRENT          (0x41)
#define ULCD_POINTER          (0x42)
#define ULCD_OPERAND0         (0x48)
#define ULCD_OPERAND1         (0x49)
#define ULCD_OPERAND2         (0x4A)
#define ULCD_OPERAND3         (0x4B)
#define ULCD_OPERAND4         (0x4C)
#define ULCD_OPERAND5         (0x4D)
#define ULCD_OPERAND6         (0x4E)
#define ULCD_OPERAND7         (0x4F)
#define ULCD_RESULT0          (0x50)
#define ULCD_RESULT1          (0x51)

// Команды ULCD
#define ULCD_REGREAD          (0x01)
#define ULCD_REGWRITE         (0x02)
#define ULCD_REGMOVE          (0x03)
#define ULCD_REGINC           (0x04)
#define ULCD_REGDEC           (0x05)
#define ULCD_REGOR            (0x06)
#define ULCD_REGAND           (0x07)
#define ULCD_BUFFERCREATE     (0x08)
#define ULCD_BUFFERLOAD       (0x0E)
#define ULCD_BUFFERSETSIZE    (0x0A)
#define ULCD_BUFFERWRITEBYTE  (0x0C)
#define ULCD_BUFFERREADBYTE   (0x0D)
#define ULCD_BUFFERGETBYTE    (0x10)
#define ULCD_BUFFERPUTBYTE    (0x11)
#define ULCD_WIDGETCREATE     (0x20)
#define ULCD_WIDGETDELETE     (0x21)
#define ULCD_SETFRAME         (0x22)
#define ULCD_WIDGETREFRESH    (0x23)
#define ULCD_WIDGETSETHANDLER (0x24)
#define ULCD_WIDGETSETCOORD   (0x26)
#define ULCD_WIDGETSETSIZE    (0x28)
#define ULCD_WIDGETSETFONT    (0x32)
#define ULCD_WIDGETSETDATA    (0x34)

#define ULCD_SOFTRESET        (0xFE)
#define ULCD_TSCALIBRATION    (0xFF)

// Параметры
#define ULCD_LOADFROMFLASH    (0)
#define ULCD_LOADFROMSD       (1)

#define ULCD_WINDOW           (0x00)
#define ULCD_FRAME            (0x01)
#define ULCD_BITMAP           (0x05)
#define ULCD_TEXT             (0x07)
#define ULCD_BUTTON           (0x0A)

// События
#define ULCD_EVT_BUTTON       (0x07)

#define ulcd_delay()          delay_ms(1)
#define ulcd_waitcmd()        while(bit_is_set(ulcd_regread(ULCD_STATUS), 0))

uint32_t ulcd_regread(uint8_t index);
void ulcd_regwrite(uint8_t index, uint32_t value);
void ulcd_regmov(uint8_t to, uint8_t from);
void ulcd_cmd(uint8_t index, uint8_t *param, uint32_t count);



int main(){
  init(); // Инициализация переферийных устройств STM32

  delay_ms(500); // Инициализация ULCD
  ulcd_cmd(ULCD_SOFTRESET, 0, 0); // Программный сброс
  delay_ms(500);
  
  uint32_t start = ulcd_regread(0x01);
  uint32_t stop = ulcd_regread(0x02);
  uint32_t coins = ulcd_regread(0x03);
  uint32_t earth = ulcd_regread(0x04);
  uint32_t bitmap = ulcd_regread(0x06);
  uint32_t button = ulcd_regread(0x07);
  uint32_t current_bitmap = coins;

  bool ledison = false;
  led_counter = 0;

  bool show = false;

  while(true){
    if(bit_is_set(GPIOB_IDR, 7)){ // Опрос вывода запроса на прерывание
      ulcd_regwrite(ULCD_INTERRUPT, 0); // Очистка флага прерывания
      // Переключение надписи на кнопке
      ulcd_regwrite(ULCD_CURRENT, button);
      if(show){
        ulcd_regwrite(ULCD_OPERAND0, start); // Метка
      }
      else {
        ulcd_regwrite(ULCD_OPERAND0, stop); // Метка
      }
      // Установка метки
      ulcd_cmd(ULCD_WIDGETSETDATA, 0, 0);
      ulcd_waitcmd();
      // Перерисовка компонента
      ulcd_cmd(ULCD_WIDGETREFRESH, 0, 0);
      ulcd_waitcmd();
      show = !show;
    }

    // Моргание светодиодом
    if(led_counter > 500){
      if(ledison){
        GPIOC_BSRR = _BV(9);
        if(show){
          // Установка изображения
          ulcd_regwrite(ULCD_CURRENT, bitmap);
          if(current_bitmap == coins){
            current_bitmap = earth;
            ulcd_regwrite(ULCD_OPERAND0, earth);
          }
          else {
            current_bitmap = coins;
            ulcd_regwrite(ULCD_OPERAND0, coins);
          }
          // Устанавливаем изображение
          ulcd_cmd(ULCD_WIDGETSETDATA, 0, 0);
          ulcd_waitcmd();
          // Перерисовка компонента
          ulcd_cmd(ULCD_WIDGETREFRESH, 0, 0);
          ulcd_waitcmd();
        }
      }
      else {
        GPIOC_BRR = _BV(9);
      }
      ledison = !ledison;
      led_counter = 0;
    }
  }
}

// Прерывание системного таймера (SysTick)
void __SysTick(){
  counter++;
  led_counter++;
}

uint8_t spi_com(uint8_t data){
  GPIOD_BRR = _BV(7);
  while(bit_is_clear(SPI3_SR, SPI3_SR_TXE));
  SPI3_DR = data;
  while(bit_is_clear(SPI3_SR, SPI3_SR_RXNE));
  data = SPI3_DR;
  GPIOD_BSRR = _BV(7);
  return data;
}

uint32_t ulcd_regread(uint8_t index){
  spi_com(ULCD_REGREAD);
  ulcd_delay();
  spi_com(index);
  ulcd_delay();
  uint32_t result = 0;
  for(uint32_t i = 0; i < 4; i++){
    uint32_t data = (uint32_t)spi_com(0x00);
    result |= (data << (8 * i));
    ulcd_delay();
  }
  return result;
}

void ulcd_regwrite(uint8_t index, uint32_t value){
  spi_com(ULCD_REGWRITE);
  ulcd_delay();
  spi_com(index);
  ulcd_delay();
  for(uint32_t i = 0; i < 4; i++){
    spi_com((value >> (8 * i)) & 0xFF);
    ulcd_delay();
  }
}

void ulcd_regmov(uint8_t to, uint8_t from){
  spi_com(ULCD_REGMOVE);
  ulcd_delay();
  spi_com(to);
  ulcd_delay();
  spi_com(from);
  ulcd_delay();
}

void ulcd_cmd(uint8_t index, uint8_t *param, uint32_t count){
  spi_com(index);
  ulcd_delay();
  for(uint32_t i = 0; i < count; i++){
    spi_com(param[i]);
    ulcd_delay();
  }
}

// Инициализация переферийных устройств
static void init(){
  // Инициализация системного таймера
  SysTickReload = TIMRELOAD_CALC(1000, 72000000); // 1кГц
  SysTickCurrent = 0;
  SysTickControl = _BV(SysTickControl_ENABLE) | _BV(SysTickControl_TICKINT) |
  _BV(SysTickControl_CLKSOURCE); // Источник тактирования -- частота
  // процессора (AHB)

  RCC_APB2ENR |= _BV(RCC_APB2ENR_AFIOEN) | _BV(RCC_APB2ENR_IOPAEN) |
  _BV(RCC_APB2ENR_IOPBEN) | _BV(RCC_APB2ENR_IOPCEN) | _BV(RCC_APB2ENR_IOPDEN);
  
  // ULCD запрос на прерывание
  GPIOB_CRL = (GPIOB_CRL & ~((3 << GPIOB_CRL_CNF7) | (3 << GPIOB_CRL_MODE7))) |
  ((1 << GPIOB_CRL_CNF7) | (0 << GPIOB_CRL_MODE7)); // PB7 -- плавающий вход

  // Конфигурация выводов SPI и ULCD

  GPIOB_CRL = (GPIOB_CRL & ~((3 << GPIOB_CRL_CNF7) | (3 << GPIOB_CRL_MODE7))) |
  ((1 << GPIOB_CRL_CNF7) | (0 << GPIOB_CRL_MODE7)); // PB7, INT -- Плавающий
  // вход

  GPIOD_CRL = (GPIOD_CRL & ~((3 << GPIOD_CRL_CNF7) | (3 << GPIOD_CRL_MODE7))) |
  ((0 << GPIOD_CRL_CNF7) | (3 << GPIOD_CRL_MODE7)); // PD7, SSEL --
  // Двухтактный выход, 50МГц

  GPIOC_CRH = (GPIOC_CRH & ~((3 << GPIOC_CRH_CNF10) |
  (3 << GPIOC_CRH_MODE10))) | ((2 << GPIOC_CRH_CNF10) |
  (3 << GPIOC_CRH_MODE10)); // PC10, SCK -- Альтернативный двухтактный выход

  GPIOC_BSRR = _BV(11);
  GPIOC_CRH = (GPIOC_CRH & ~((3 << GPIOC_CRH_CNF11) |
  (3 << GPIOC_CRH_MODE11))) | ((2 << GPIOC_CRH_CNF11) |
  (0 << GPIOC_CRH_MODE11)); // PC11, MISO -- Подтянутый к питанию вход

  GPIOC_CRH = (GPIOC_CRH & ~((3 << GPIOC_CRH_CNF12) |
  (3 << GPIOC_CRH_MODE12))) | ((2 << GPIOC_CRH_CNF12) |
  (3 << GPIOC_CRH_MODE12)); // PC12, MOSI -- Альтернативный двухтактный выход

  AFIO_MAPR |= (1 << AFIO_MAPR_SPI3_REMAP);

  // Инициализация SPI
  RCC_APB1ENR |= _BV(RCC_APB1ENR_SPI3EN);
  SPI3_CR1 = (1 << SPI3_CR1_CPHA) | (1 << SPI3_CR1_CPOL) |
  (1 << SPI3_CR1_MSTR) | (1 << SPI3_CR1_BR) | (1 << SPI3_CR1_SPE) |
  (0 << SPI3_CR1_LSBFIRST) | (0 << SPI3_CR1_DFF) | (1 << SPI3_CR1_SSM) |
  (1 << SPI3_CR1_SSI);

  GPIOD_BSRR = _BV(7); // SSEL = 1

  // Инициализация светодиода (PC9)
  GPIOC_CRH = (GPIOC_CRH & ~(0xF << 4)) | (0x1 << 4); // Двухтактный
  // выход, 10МГц
  GPIOC_BRR = _BV(9); // PC9 = 0
}

void memclr(void *buffer, size_t size){
  for(uint32_t i = 0; i < size; i++){
    *((uint8_t *)buffer++) = 0;
  }
}

